#!/usr/bin/env python

from __future__ import print_function  # Imports print_function from python 3

import os
import re
import argparse
import requests
import requests_ftp
import sys
import subprocess
import shlex
import time
import socket

if (sys.version_info > (3, 0)):
    from urllib.parse import urlparse
else:
    from urlparse import urlparse

from multiprocessing import Pool
from functools import partial


__version__ = "1.1.1"


class distfiles(): 
    def __init__(self, directory):
        self.author = "Emanuel Haupt <ehaupt@FreeBSD.org>"
        self.directory = directory

        if not os.path.isdir(self.directory):
            print(self.directory, "Directory does not exist.")
            sys.exit()

    def __get_dist_urls__(self):
        make_cmd = 'make -C %s fetch-urlall-list DISTDIR=%s MASTER_SITE_BACKUP=""' %  ( self.directory, "/tmp" )
        proc = subprocess.Popen(shlex.split(make_cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = proc.communicate()

        if err:
            print(self.directory, "does not seem to be a valid port.")
            sys.exit(2)

        if out:
            urls = out.decode('utf-8').split()
            return(urls)

    def __get_www_urls__(self):
        make_cmd = 'make -C %s describe DISTDIR=%s MASTER_SITE_BACKUP=""' %  ( self.directory, "/tmp" )
        proc = subprocess.Popen(shlex.split(make_cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = proc.communicate()

        if err:
            print(self.directory, "does not seem to be a valid port.")
            sys.exit(2)

        if out:
            wwws = out.decode('utf-8').split('|')[-1].rstrip("\n")
            if not wwws == "":
                return([wwws])
            else:
                return(None)

    def get_all_urls(self):
        url = {}
        dist=self.__get_dist_urls__()
        www=self.__get_www_urls__()

        if not dist is None:
            url['dist_urls'] = dist

        if not www is None:
            url['www_urls'] = www

        return(url)


def hostname_resolves(hostname):
    try:
        socket.gethostbyname(hostname)
        return True
    except socket.error:
        return False

def resolve_or_timeout_error(url, urltype, args):
    tmout_code = "500"
    nxdomain_code = '500'
    hostname=urlparse(url).hostname
    if hostname_resolves(hostname):
        print("%s [%s]\t%s (timeout: %s)" % (tmout_code, urltype, url, args.timeout))
    else:
        # host does not resolve
        print("%s [%s]\t%s (%s does not resolve)" % (nxdomain_code, urltype, url, hostname))


def validate_url_pool(args, iterable):
    for url, urltype in iterable.items():
        if re.match(r'^http', url):
            try:
                r = requests.head(url, timeout=args.timeout)

                if r.is_redirect:
                    print("%s [%s]\t%s -> %s" % (r.status_code, urltype, url, r.headers['Location']))
                else:
                    if r.status_code >= 200 and r.status_code < 300 and args.silent:
                        continue

                    print("%s [%s]\t%s" % (r.status_code, urltype, url))

            except:
                resolve_or_timeout_error(url, urltype, args)

        elif re.match(r'^ftp', url):
            try:
                requests_ftp.monkeypatch_session()
                s = requests.Session()
                r = s.head(url, timeout=args.timeout)
                status_code = r.status_code
                if r.status_code >= 200 and r.status_code < 300 and args.silent:
                    continue

                print("%s [%s]\t%s" % (status_code, urltype, url))

            except:
                resolve_or_timeout_error(url, urltype, args)

        else:
            print("%s not supported" % url)


def main():
    parser = argparse.ArgumentParser(
        description = 'Script to check the availability of a ports distfiles',
        epilog = 'For further testing, please consult https://www.freebsd.org/doc/en_US.ISO8859-1/books/porters-handbook/'
    )

    parser.add_argument('directory', nargs='?', default=os.getcwd(),
                        help="ports directory")
    parser.add_argument('-j', '--jobs', type=int, default=30,
                        help='number of validation threads (default 30)')
    parser.add_argument('-s', '--silent', action='store_true', default=False,
                        help='silent mode, only print errors')
    parser.add_argument('-t', '--timeout', type=float, default=5, 
                        help='link check timeout in seconds (default 5)')
    parser.add_argument('-v', '--version', action='store_true', help='print version')

    args = parser.parse_args()

    dist = distfiles(args.directory)

    if args.version:
        print("distilator version", __version__)
        sys.exit()

    url=dist.get_all_urls()

    iterable = []
    for urltype, urls in url.items():
        for url in urls:
            if urltype == "dist_urls":
                iterable.append({ url : 'DISTFILE' })
            elif urltype == "www_urls":
                iterable.append({ url : 'WWW' })
            else:
                iterable.append({ url : 'UNKNOWN' })

    pool = Pool(processes=args.jobs)
    func = partial(validate_url_pool, args)
    pool.map(func, iterable)
    pool.close()
    pool.join()


# Main section of the script
if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:  # Catch Ctrl-C
        sys.exit(0)
    except SystemExit as sysexit:
        if not sysexit.code == 0:
            raise
        else:
            sys.exit(0)
    except:
        raise
